//
//	Kernel Memory usermode support
//
//	www.catch22.net
//
#include <windows.h>
#include <stdio.h>
#include <aclapi.h>
#include "kmem.h"

static HANDLE			g_hSection;
static PHARDWARE_PTE	g_PDE;
static LONG				g_RefCount = 0;

ULONG kmemrawcpy(ULONG physaddr, PVOID buf, ULONG len);

static ULONG CurPde() 
{
    unsigned short tr; 
	LDT_ENTRY tss;
	ULONG	virtaddr;
	ULONG	physaddr;
	ULONG	cr3;
	DWORD	affinity;

	__asm str tr

	affinity = SetThreadAffinityMask(GetCurrentThread(), 1);
	
    GetThreadSelectorEntry(GetCurrentThread(), tr, &tss);
	
	SetThreadAffinityMask(GetCurrentThread(), affinity);

    virtaddr = (tss.HighWord.Bits.BaseHi << 24) + (tss.HighWord.Bits.BaseMid << 16) + tss.BaseLow;

    physaddr = virtaddr & ~0x80000000;
	
    kmemrawcpy(physaddr + 0x1C, &cr3, sizeof(cr3));
	
    return (cr3 & 0xFFF) == 0 ? cr3 : 0;
}

 
static ULONG AnyPde() 
{
    const ULONG reserved = 0xA0;
	
    PULONG lowmem;
	
    ULONG pfn;
	
	lowmem = (PULONG)MapViewOfFile(g_hSection, FILE_MAP_READ, 0, 
		0, reserved*0x1000);

	//lowmem = (PULONG)MapViewOfFile(g_hSection, FILE_MAP_READ, 0, 
	//	0x20000-1, reserved*0x1000);

	if(lowmem == 0)
	{
		printf("Failed: %x\n", GetLastError());
		return 0;
	}

	if(lowmem == 0)
		return 0;
	
    for(pfn = 0; pfn < reserved; pfn++)
	{
		if (lowmem[pfn * 0x400 + 0x300] == (pfn+31) * 0x1000 + 0x67)
		{
			pfn += 31;
			UnmapViewOfFile(lowmem);
			return pfn * 0x1000;
		}
	}
	
	UnmapViewOfFile(lowmem);
	
	return 0;
}

 
static ULONG GetPde()
{
    ULONG pde = CurPde();
	
    return pde == 0 ? AnyPde() : pde;
	//return 0x39000;
}

static PVOID MapSystemPage(ULONG virtaddr, ULONG prot) 
{
    ULONG	 	 offset;
	HARDWARE_PTE pte1;
	 
	pte1 = g_PDE[virtaddr >> 22];
	
    if(!pte1.Valid) 
	{
		return 0;
	}
	
    if(pte1.LargePage == 0) 
	{
        PHARDWARE_PTE table;
        HARDWARE_PTE  pte2;

		table = (PHARDWARE_PTE)MapViewOfFile(g_hSection, FILE_MAP_READ, 0, pte1.PageFrameNumber << 12, 0x1000);
		pte2  = table[(virtaddr >> 12) & 0x3ff];
		
        UnmapViewOfFile(table);
		
        if(!pte2.Valid)
		{
			return 0;
		}
		
        offset = pte2.PageFrameNumber << 12;
    }
    else
	{
        offset = (pte1.PageFrameNumber << 12) + (virtaddr & 0x3ff000);
	}
	
	return MapViewOfFile(g_hSection, prot, 0, offset, PAGE_SIZE);
}

 
ULONG kmemcpy(PVOID dest, PVOID source, ULONG len) 
{
    ULONG m = 0, copied = 0;
	ULONG y, x;
	ULONG virtaddr = (ULONG)source;
	ULONG buf      = (ULONG)dest;

    for(x = virtaddr; x < virtaddr + len; x += y) 
	{
        PVOID q = MapSystemPage(x, FILE_MAP_READ);
		
		y = min(len - m, PAGE_SIZE - BYTE_OFFSET(x));

        if (q != 0)
		{
			memcpy((PVOID)(buf + m), (PBYTE)q + BYTE_OFFSET(x), y);
		
			UnmapViewOfFile(q);

			copied += y;
		}
		else
		{
		//	printf("failed MapSystemPage\n");
			memset((PVOID)(buf + m), 0, y);

		}
		
		m += y;
    }
	
    return copied;
}

 
ULONG kmemrawcpy(ULONG physaddr, PVOID buf, ULONG len) 
{
    ULONG m = 0;
	ULONG y, x;

    for(x = physaddr; x < physaddr + len; x += y) 
	{
        PVOID q = MapViewOfFile(g_hSection, FILE_MAP_READ, 0, x & ~0xFFF, 0x1000);
		
        if (q == 0) return m;
		
        y = min(len - m, PAGE_SIZE - BYTE_OFFSET(x));
		
        memcpy((PBYTE)buf + m, (PBYTE)q + BYTE_OFFSET(x), y);
		
        UnmapViewOfFile(q);
		
        m += y;
    }
	
    return m;
}

void KmInit()
{
    if(InterlockedIncrement(&g_RefCount) == 1)
	{
		UNICODE_STRING			name; 
		OBJECT_ATTRIBUTES		oa = { sizeof(oa), 0, &name, OBJ_CASE_INSENSITIVE};
		
		ZWOPENSECTION			ZwOpenSection;
		RTLINITUNICODESTRING	RtlInitUnicodeString;
		
		HMODULE					hNtDll = GetModuleHandle("ntdll.dll");
		
		//
		//	Get addresses of native APIs
		//
		RtlInitUnicodeString = (RTLINITUNICODESTRING) GetProcAddress(hNtDll, "RtlInitUnicodeString");
		ZwOpenSection        = (ZWOPENSECTION)		  GetProcAddress(hNtDll, "ZwOpenSection");
		
		//
		//	Open \Device\PhysicalMemory and map a section to it
		//
		RtlInitUnicodeString(&name, L"\\Device\\PhysicalMemory");
		ZwOpenSection(&g_hSection, SECTION_MAP_READ, &oa);

		g_PDE = (PHARDWARE_PTE)MapViewOfFile(g_hSection, FILE_MAP_READ, 0, GetPde(), 0x1000);
	}
}


void KmDeInit()
{
	if(InterlockedDecrement(&g_RefCount) == 0)
	{
		UnmapViewOfFile(g_PDE);
		CloseHandle(g_hSection);
	}
}